//------------------------------------------------------------------------
//
// Copyright (C) 1997-2009 iNTERFACEWARE Inc.  All Rights Reserved
//
// Module:  C++ Destination Plugin Example
// 
// Description
//
// This application is a simple example of using the Chameleon generated 
// objects within an Iguana plugin.  It is intended to be a helpful starting 
// point in understanding how to use these objects but must be fleshed out 
// in the following areas to provide a full solution for an application 
// HL7 interface:
//
// 1) The code will need to be modified to insert the parsed data into a 
//    database whether by using explicit database calls or going through 
//    some type of middle tier layer that your application may support.
//
// 2) Definitely take the time to read the manual and go through the 
//    tutorials after looking at this example before attempting to deploy 
//    a production solution.  We do provide custom on-site training and 
//    technical implementation services if you require them.
//
// Author: iNTERFACEWARE
// Date:   Wed 08/17/2009
//---------------------------------------------------------------------------
#include "CHMserverPreCompHead.h"
#pragma hdrstop

#include <IGC/IGCdestination.h>
#include <IGC/IGCdestinationMessageReceivedEventInputs.h>
#include <IGC/IGCdestinationMessageReceivedEventOutputs.h>
#include <IGC/IGCdisconnectReason.h>
#include <IGC/IGCerrorClass.h>

#include <CHM/CHMerrorClass.h>
#include <CHM/CHMstring.h>
#include <CHM/CHMutil.h>

#include <iostream>
using namespace std;

/* template code */

class ToPluginExample : public IGCdestination
{
public:
   // Simple logging class: outputs both to stdout and a string.
   class OutputLog : public CHMoutputLog
   {
   public:
      virtual void OutputLine(const CHMchar* Line)
      {
         m_LogString.append(Line);
         m_LogString.append("\n");

         cout << Line << endl;
      }

      CHMstring m_LogString;
   };


   ToPluginExample() : IGCdestination(), m_pEngineOutputLog(0)
   {
      cout << "Trying to load test_data.vmd." << endl;

      m_Engine.Load("test_data.vmd");
      m_Engine.SetLog(m_pEngineOutputLog = new OutputLog);
      m_Engine.SetCurrentConfigurationIndex(0);

      cout << "Loaded message definition file test_data.vmd." << endl;
   }

   ~ToPluginExample()
   {
   }

   PrefixTestappEngine m_Engine;
   OutputLog* m_pEngineOutputLog; // Owned (deleted) by m_Engine

   // Message processing invoked by OnMessageReceived and 
   // OnResubmissionPreviewMessageReceived event handlers.
   void ProcessMessage
   (
      IGCdestinationMessageReceivedEventInputs& Inputs,
      IGCdestinationMessageReceivedEventOutputs& Outputs,
      bool ResubmissionPreview
   )
   {
      // The ResubmissionPreview flag indicates whether or not we're processing a resubmission 
      // preview message.  When processing a resubmission preview message, the plugin should 
      // simulate the processing that would happen for a regular message, but without causing 
      // external side effects.  For example, no database commits should be made for resubmission
      // preview messages.  In this example, regular message processing has no external side 
      // effects, so we simply ignore the ResubmissionPreview flag.

      // Message processing goes here. This example just prints the received message and
      // uses the Chameleon engine to parse the message.
      cout << endl;
      cout << "*** Message Received ***" << endl;
      cout << CHMlineFeedConvert(Inputs.Message()).c_str() << endl << endl;

      try
      {
         // We manually invoke the engine to parse the message.  The second
         // parameter is set to 'null'; it used to be part of the transport framework for
         // Chameleon, but is no longer required.
         m_pEngineOutputLog->m_LogString.clear();
         m_Engine.ParseMessage(Inputs.Message(), 0);

         // Tell Iguana that the message has been successfully processed.
         Outputs.SetMessageProcessedFlag( true );
      }
      catch(const CHMerror& Error)
      {
         // Tell Iguana that the message was not successfully processed.
         Outputs.SetMessageProcessedFlag( false );

         // Add an error-level log message to Iguana's logs describing the failure.
         Outputs.AddErrorLogMessage(Error.Description());

         // You can use ResendMessageFlag and ResendMessageDelay to request that Iguana
         // resend the same message again after a delay period.  This is useful in situations
         // where the message processing failure was due to a time sensitive cause, and
         // message processing might succeed at a later time.
         //
         // Outputs.SetResendMessageFlag( true );
         // Outputs.SetResendMessageDelay( 1000 ); // Resend after 1 second

         // You can use StopChannel, StopChannelErrorFlag, and StopChannelReason to stop the Iguana 
         // channel with or without an error flag, and provide a reason for the stop request.
         //
         // Outputs.SetStopChannel(true);
         // Outputs.SetStopChannelErrorFlag(true);
         // Outputs.SetStopChannelReason("An error occurred.");
      }

      // You can send additional error/warning/info/debug log messages to Iguana.  In this case
      // we simply send an info-level log message containing the Chameleon engine's log output.
      if (!m_pEngineOutputLog->m_LogString.is_null())
      {
         Outputs.AddInfoLogMessage(m_pEngineOutputLog->m_LogString.c_str());
      }
   }

   // This event is triggered when a message is received.
   virtual void OnMessageReceived
   (
      IGCdestinationMessageReceivedEventInputs& Inputs,
      IGCdestinationMessageReceivedEventOutputs& Outputs
   )
   {
      // Note that exceptions can never be thrown outside the handler.
      try
      {
         ProcessMessage(Inputs, Outputs, false);
      }
      catch (const IGCerror& Error)
      {
         cout << "Error: " << Error.Description() << endl;
      }
      catch (...)
      {
         cout << "Error: unexpected exception" << endl;
      }
   }

   // This event is triggered when a resubmission preview message is received.
   virtual void OnResubmissionPreviewMessageReceived
   (
      IGCdestinationMessageReceivedEventInputs& Inputs,
      IGCdestinationMessageReceivedEventOutputs& Outputs
   )
   {
      // Note that exceptions can never be thrown outside the handler.
      try
      {
         // This handler should simulate the processing that would be performed by the
         // OnMessageReceived event handler without causing external side effects.
         // For example, database changes should not be made for resubmission preview messages.

         ProcessMessage(Inputs, Outputs, true);
      }
      catch (const IGCerror& Error)
      {
         cout << "Error: " << Error.Description() << endl;
      }
      catch (...)
      {
         cout << "Error: unexpected exception" << endl;
      }
   }

   // This event is triggered when connection to Iguana is closed.
   // See also the use of WaitForDisconnect() below.
   virtual void OnDisconnected()
   {
      // Note that exceptions can never be thrown outside the handler.
      try
      {
         cout << "Destination disconnected." << endl;
      }
      catch (...)
      {
         cout << "Error: unexpected exception" << endl;
      }
   }
};

int main(int argc, const char** argv)
{
   try
   {
      // Disable buffering for output streams, so that Iguana receives output immediately
      // rather than when the output buffer happens to be filled.  Note that this function
      // must be called before any output is written.  An alternative to this approach
      // would be to call fflush() judiciously after all output commands.
      IGC_DISABLE_OUTPUT_BUFFERING();

      if (argc < 4)
      {
         throw IGCerror("Invalid command line arguments.", __LINE__, __FILE__);
      }

      const char* pHostname    = argv[1];
      int         Port         = atoi( argv[2] );
      const char* pChannel     = argv[3];

      // Connect to Iguana.
      ToPluginExample Destination;
      cout << "Connecting to Iguana" << endl;
      Destination.Connect(pHostname, Port, pChannel);

      // Wait for the plugin to disconnect.
      // Could alternatively use WaitForDisconnectWithTimeout() here.
      Destination.WaitForDisconnect();

      // Print the reason for disconnection.
      IGCdisconnectReason DisconnectReason = Destination.DisconnectReason();
      cout << DisconnectReason.Description() << endl;

      return 0;
   }
   catch (const IGCerror& Error)
   {
      cout << "Error: " << Error.Description() << endl << endl;
   }
   catch (const CHMerror& Error)
   {
      cout << "Error: " << Error.Description() << endl << endl;
   }
   catch (...)
   {
      cout << "Error: unexpected exception" << endl << endl;
   }

   cout << "This is an Iguana destination plugin.  It runs as a command line application\n"
           "which connects via TCP/IP to the plugin communication server within Iguana.\n"
           "It will be necessary to run a channel in Iguana and set up a channel with\n"
           "a To Plugin component for this to connect to.  See:\n"
           "\n"
           "   http://www.interfaceware.com/manual/plugin_setup_cpp.html\n"
           "\n"
           "Once the channel is set up and running you can connect to it with this usage:\n"
           "\n"
           "   " << argv[0] << " <iguana hostname> <plugin communication port> <channel name>\n"
           "\n"
           "The plugin communication port is the port that Iguana listens on for plugin\n"
           "communication." << endl;

   return EXIT_FAILURE;
}

